/*
 * Copyright (C) 2000-2002 HighPoint Technologies, Inc.
 * HPT3xx UDMA/ATA RAID Controller Device Driver for Linux
 * File:	hpt.c
 */
#define HPT_DEFINE_VERSION
#include "hptkern.h"
#include "hptglb.h"
#include "hpt.h"

#ifdef DEBUG
#ifdef DEBUG_LEVEL
int hpt_dbg_level = DEBUG_LEVEL;
#else
int hpt_dbg_level = 0;
#endif
#endif

/*
 * SMP kernels will hang when create thread in Detect()...
 */
#if __BOOT_KERNEL_SMP
#define START_WORKER_THREAD_BEFORE_INIT
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#define GLOBAL_DATA_ALLOC(size) ({ void *p=kmalloc(size, GFP_ATOMIC); if (p) memset(p, 0, size); p; })
#define GLOBAL_DATA_FREE(p, size) kfree(p)
#else
#define GLOBAL_DATA_ALLOC(size) scsi_init_malloc(size, GFP_ATOMIC)
#define GLOBAL_DATA_FREE(p, size) scsi_init_free((char*)p, size)
#endif

struct Scsi_Host *hpt_vhost = 0;
Scsi_Cmnd *hpt3xx_req_last = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
/* notifier block to get a notify on system shutdown/halt/reboot */
static int hpt_halt(struct notifier_block *nb, ulong event, void *buf);
static struct notifier_block hpt_notifier = {
    hpt_halt, NULL, 0
};
#endif

static spinlock_t hptio_request_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t gio_request_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t queue_request_lock = SPIN_LOCK_UNLOCKED;

/****************************************************************
 *	Name:	Irq_Handler	:LOCAL
 *	Description:	Interrupt handler.
 *	Parameters:		irq		- Hardware IRQ number.
 *					dev_id	-
 *					regs	-
 *	Returns:		TRUE if drive is not ready in time.
 ****************************************************************/
static void Irq_Handler (int irq, void *dev_id, struct pt_regs *regs)
{
    unsigned long		flags;

	KdPrintI(("----- Entering Isr() -----\n"));
    spin_lock_irqsave (&io_request_lock, flags);

	hpt_handler(irq, dev_id);

    spin_unlock_irqrestore (&io_request_lock, flags);
	KdPrintI(("----- Leaving Isr() -----\n"));
}

/****************************************************************
 *	Name:	hpt3xx_QueueCommand
 *	Description:	Process a queued command from the SCSI manager.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *					done  - Pointer to done function to call.
 *	Returns:		Status code.
 ****************************************************************/
int hpt3xx_QueueCommand (Scsi_Cmnd *SCpnt, void (*done)(Scsi_Cmnd *))
{
    unsigned long flags;

    KdPrintI(("hpt3xx_QueueCommand() %d/%d/%d cdb=(%x-%x-%x)\n",
        SCpnt->channel, SCpnt->target, SCpnt->lun,
			*((PULONG)&SCpnt->cmnd), *((PULONG)&SCpnt->cmnd+1), *((PULONG)&SCpnt->cmnd+2)));
    SCpnt->scsi_done = (void *)done;
    
	spin_lock_irqsave(&queue_request_lock, flags);
	if(hpt3xx_req_last==NULL)
		SCpnt->SCp.ptr = (char*)SCpnt;
	else
	{
		SCpnt->SCp.ptr = hpt3xx_req_last->SCp.ptr;
		hpt3xx_req_last->SCp.ptr = (char*)SCpnt;
	}
	hpt3xx_req_last = SCpnt;
	spin_unlock_irqrestore (&queue_request_lock, flags);	
       
	if(Driver_Free()) {
		spin_lock_irqsave(&gio_request_lock, flags);
		TaskQueue();
		spin_unlock_irqrestore (&gio_request_lock, flags);	
		}

	KdPrintI(("----- Leaving hpt3xx_queueCommand() -----\n"));
	return 0;
}

/****************************************************************
 *	Name:	internal_done :LOCAL
 *	Description:	Done handler for non-queued commands
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		Nothing.
 ****************************************************************/
static void internal_done (Scsi_Cmnd * SCpnt)
{
	SCpnt->SCp.Status++;
}

/****************************************************************
 *	Name:	hpt3xx_Command
 *	Description:	Process a command from the SCSI manager.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		Status code.
 ****************************************************************/
int hpt3xx_Command (Scsi_Cmnd *SCpnt)
{
	DWORD timeout;
	
    SCpnt->SCp.Status = 0;
	hpt3xx_QueueCommand (SCpnt, internal_done);

	timeout = jiffies + 60 * HZ; // should be longer than hard-reset time
	while (!SCpnt->SCp.Status && time_before(jiffies, timeout)) barrier();

	if (!SCpnt->SCp.Status) SCpnt->result = (DID_ERROR<<16);
	return SCpnt->result;
}

static void launch_worker_thread(void)
{
	DECLARE_MUTEX_LOCKED(sem);
	hpt_notify_wt = &sem;

	kernel_thread((int (*)(void *))hpt_worker_thread, (void*)0, 0);

	down (&sem);
	hpt_notify_wt = NULL;
}

/****************************************************************
 *	Name:	hpt3xx_init
 *	Description:	initialize a adapter
 *	Returns:		0 success
 ****************************************************************/
static int hpt3xx_init(PHPT_ADAPTER pAdap, struct pci_dev *pPciDev)
{
	PChannel			pChan;
	int					z;
	int					setirq;
	UINT				BMI;
	
	/*
	 * common settings
	 */	
	pAdap->pci_bus = pPciDev->bus->number;
	pAdap->devfn = pPciDev->devfn;
	pAdap->irq = pPciDev->irq;
	pAdap->irqOwned = 0;
	
	printk("Found Controller: %s\n", pAdap->name);

	pAdap->IDEChannel = (PChannel)GLOBAL_DATA_ALLOC(sizeof(Channel)*pAdap->num_buses);
	if (pAdap->IDEChannel==0)
		return -1;

	setirq = 1;
	for (z = 0;  z < NumAdapters; z++) {	 
		if (hpt_adapters[z]->irqOwned == pAdap->irq) setirq = 0;
	}
	if (setirq) {
		if (request_irq (pAdap->irq, Irq_Handler, SA_SHIRQ, PROC_DIR_NAME, pAdap) < 0) {
			if (request_irq (pAdap->irq, Irq_Handler, SA_INTERRUPT | SA_SHIRQ, PROC_DIR_NAME, pAdap) < 0 ){
				printk ("Unable to allocate IRQ for %s\n", pAdap->name);
				return -1;
			}
		}
		pAdap->irqOwned = pAdap->irq;
	}

	/*
	 * chip specific initialization
	 */
	switch(pAdap->chip_type) {
	case CHIP_TYPE_HPT372A:
	case CHIP_TYPE_HPT372:
		pcibios_write_config_byte(pAdap->pci_bus, pAdap->devfn, 0x70, 0);
		/* fall through */
	case CHIP_TYPE_HPT370:
		pcibios_read_config_dword(pAdap->pci_bus, pAdap->devfn, REG_BMIBA, &BMI);
		BMI &= ~1;
		/* this will set both primary and secondary channel. */
		SetHptChip(pAdap->IDEChannel, (PUCHAR)BMI);
		break;
	default:
		return -1;
	}

	/*
	 * initialize channels
	 */
	for (z=0; z<pAdap->num_buses; z++ ){
		/*
		 * Set Scat Gath List buffer for the channel. Must be 16 byte align
		 */
		pChan = &pAdap->IDEChannel[z];
		pChan->SgPhysicalAddr = (MapToPhysAddr((PVOID)pChan->SGBuf) + 15) & 0xfffffff0;
		pChan->pSgTable = (PSCAT_GATH)MapToLogicAddr(pChan->SgPhysicalAddr);

		AtapiHwInitialize(pChan);

		pChan->busy_timer = (struct timer_list *)GLOBAL_DATA_ALLOC(sizeof(struct timer_list)*3);
		if (pChan->busy_timer == 0)
			return -1;
		init_timer(pChan->busy_timer);
		pChan->busy_timer->function = (void (*)(ULONG))hpt_busy_timer;
		pChan->busy_timer->data = (ULONG)pChan;

		pChan->retry_timer = pChan->busy_timer +1;
		init_timer(pChan->retry_timer);
		pChan->retry_timer->function = (void (*)(ULONG))hptretry;
		pChan->retry_timer->data = (ULONG)pChan;

		pChan->watch_timer = pChan->busy_timer +2;
		init_timer(pChan->watch_timer);
		pChan->watch_timer->function = (void (*)(ULONG))hpt_watch_timer;
		pChan->watch_timer->data = (ULONG)pChan;

		pChan->adjusted = 0;
	}

	hpt3xx_init_post(pAdap);
	return 0;
}

/****************************************************************
 *	Name:	hpt3xx_Detect
 *	Description:	Detect and initialize our boards.
 *	Parameters:		tpnt - Pointer to SCSI host template structure.
 *	Returns:		Number of adapters installed.
 ****************************************************************/
int hpt3xx_Detect (Scsi_Host_Template *tpnt)
{
	PHPT_ADAPTER pAdap;
	struct pci_dev *pPciDev;
	UCHAR rev;

	printk("%s\nVersion %s, Compiled %s %s\n", 
		DRIVER_NAME, DRIVER_VERSION,  __DATE__, __TIME__);
	/* 
	 * calculate page_offset.
	 */
	hpt_page_offset = (ULONG)&hpt_adapters - virt_to_phys(&hpt_adapters);

	if (!pci_present()) {
		printk("PCI BIOS not present\n");
		return 0;
	}
	
	/* 
	 * search for HPT370 chip 
	 */
	pPciDev = NULL;
	while ((pPciDev=pci_find_device(HPT370_VENDORID, HPT370_DEVICEID, pPciDev))){
		if (NumAdapters>=MAX_ADAPTERS) break;
		pAdap = (PHPT_ADAPTER)GLOBAL_DATA_ALLOC(sizeof(HPT_ADAPTER));
		pci_read_config_byte(pPciDev, REG_RID, &rev);
		if (rev<5) {
			pAdap->chip_type = CHIP_TYPE_HPT370;
			pAdap->name = CONTROLLER_NAME_HPT370;
		}
		else {
			pAdap->chip_type = CHIP_TYPE_HPT372;
			pAdap->name = CONTROLLER_NAME_HPT372;
		}
		pAdap->num_buses = 2;
		if (hpt3xx_init(pAdap, pPciDev)==0)
			hpt_adapters[NumAdapters++] = pAdap;
	}
	/* 
	 * search for HPT372A chip 
	 */
	pPciDev = NULL;
	while ((pPciDev=pci_find_device(HPT372A_VENDORID, HPT372A_DEVICEID, pPciDev))){
		if (NumAdapters>=MAX_ADAPTERS) break;
		pAdap = (PHPT_ADAPTER)GLOBAL_DATA_ALLOC(sizeof(HPT_ADAPTER));
		pAdap->chip_type = CHIP_TYPE_HPT372A;
		pAdap->name = CONTROLLER_NAME_HPT372A;
		pAdap->num_buses = 2;
		if (hpt3xx_init(pAdap, pPciDev)==0)
			hpt_adapters[NumAdapters++] = pAdap;
	}
	
	if (NumAdapters) {
		/*
		 * register a virtual host.
		 */
		hpt_vhost = scsi_register(tpnt, 0);
		hpt_vhost->io_port		= 0;
		hpt_vhost->n_io_port	= 0;
		hpt_vhost->max_channel	= 0;  /* channel <= max_channel */
		hpt_vhost->max_id		= MAX_DEVICES;  /* id < max_id */
		hpt_vhost->max_lun     = 1;  /* lun < max_lun */
		
		Final_Array_Check();
		Set_Logical_Drives();

		/* launch worker thread */
#ifndef START_WORKER_THREAD_BEFORE_INIT
		launch_worker_thread();
#endif

#ifndef __BOOT_KERNEL_BOOT
#ifdef DRIVER_REBUILD
		{
			PVirtualDevice pArray;
			for (pArray=VirtualDevices; pArray<pLastVD; pArray++) {
				if (pArray->RaidFlags & RAID_FLAGS_NEED_SYNCHRONIZE)
					hpt_queue_dpc((HPT_DPC)hpt_rebuild_mirror, pArray);
			}
		}
#endif
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
        register_reboot_notifier(&hpt_notifier);
#endif

		return 1;
	}
	return 0;
}

/****************************************************************
 *	Name:	hpt3xx_Abort
 *	Description:	Process the Abort command from the SCSI manager.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		Always snooze.
 ****************************************************************/
int hpt3xx_Abort (Scsi_Cmnd *SCpnt)
{
	return SCSI_ABORT_SNOOZE;
}

/****************************************************************
 *	Name:	hpt3xx_Reset
 *	Description:	Process the Reset command from the SCSI manager.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *					flags - Flags about the reset command
 *	Returns:		No active command at this time, so this means
 *					that each time we got some kind of response the
 *					last time through.  Tell the mid-level code to
 *					request sense information in order to decide what
 *					to do next.
 ****************************************************************/
int hpt3xx_Reset (Scsi_Cmnd *SCpnt, unsigned int reset_flags)
{
	return SCSI_RESET_PUNT;
}

/****************************************************************
 *	Name:	hpt3xx_cleanup
 *	Description:	cleanup routine
 ****************************************************************/
static void hpt3xx_cleanup(void)
{
#ifdef DRIVER_REBUILD
	{
		// set a flag to abort rebuilding process.
		PVirtualDevice pArray;
		for (pArray=VirtualDevices; pArray<pLastVD; pArray++) {
			if (pArray->RaidFlags & RAID_FLAGS_BEING_BUILT)
				pArray->RaidFlags |= RAID_FLAGS_ABORT_REBUILD;
		}
	}
#endif
	// kill kernel thread
	if (hpt_worker_thread_handle)
	{
		DECLARE_MUTEX_LOCKED(sem);
		hpt_notify_wt = &sem;
		stop_thread = 1;
		up(hpt_queue_ready);
		down(&sem);
		hpt_notify_wt = NULL;
	}
}

/****************************************************************
 *	Name:	hpt3xx_Release
 *	Description:	Release resources allocated for a single each adapter.
 *	Parameters:		pshost - Pointer to SCSI command structure.
 *	Returns:		zero.
 ****************************************************************/
int hpt3xx_Release (struct Scsi_Host *pshost)
{
    PHPT_ADAPTER pAdap;
	int i, j;
	for (i=0; i<NumAdapters; i++) {
		pAdap = hpt_adapters[i];
		if ( pAdap->irqOwned )
			free_irq (pAdap->irq, pAdap);
		for(j=0; j<pAdap->num_buses; j++)
			GLOBAL_DATA_FREE(pAdap->IDEChannel[j].busy_timer, sizeof(struct timer_list)*3);
		GLOBAL_DATA_FREE(pAdap->IDEChannel, sizeof(Channel)*pAdap->num_buses);
		GLOBAL_DATA_FREE(pAdap, sizeof(HPT_ADAPTER));
	}
    scsi_unregister(pshost);
    return 0;
}

/****************************************************************
 *	Name:	hpt3xx_BiosParam
 *	Description:	Process the biosparam request from the SCSI manager to
 *					return C/H/S data.
 *	Parameters:		disk - Pointer to SCSI disk structure.
 *					dev	 - Major/minor number from kernel.
 *					geom - Pointer to integer array to place geometry data.
 *	Returns:		zero.
 ****************************************************************/
int hpt3xx_BiosParam (Scsi_Disk *disk, kdev_t dev, int geom[])
{
	int heads, sectors, cylinders;
	
	/* Default heads (64) & sectors (32) */
	/* Handle extended translation size for logical drives > 1Gb */
	if (disk->capacity >= 0x200000) {
		heads = 255;
		sectors = 63;
	}
	else {
		heads = 64;
		sectors = 32;
	}
	cylinders = disk->capacity / (heads * sectors);
	
	/* return result */
	geom[0] = heads;
	geom[1] = sectors;
	geom[2] = cylinders;
	
	return 0;
}

/****************************************************************
 *	Name:	hpt_halt
 *	Description:	shutdown routine
 ****************************************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
static int hpt_halt(struct notifier_block *nb, ulong event, void *buf)
{
	if (event != SYS_RESTART && event != SYS_HALT && event != SYS_POWER_OFF)
		return NOTIFY_DONE;
		
	hpt3xx_cleanup();
		
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,22)
	unregister_reboot_notifier(&hpt_notifier);
#endif
    return NOTIFY_OK;
}
#endif

static void hpt_scsi_unregister_module(int module_type, void *ptr)
{
	hpt3xx_cleanup();
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
		unregister_reboot_notifier(&hpt_notifier);
#endif
	scsi_unregister_module(module_type, ptr);
}

#ifdef START_WORKER_THREAD_BEFORE_INIT
#define init_module scsi_init_module
#endif

#undef scsi_unregister_module
#define scsi_unregister_module hpt_scsi_unregister_module
static Scsi_Host_Template driver_template = HPT3xx;
#include "scsi_module.c"

#ifdef START_WORKER_THREAD_BEFORE_INIT
#undef init_module
int init_module(void)
{
	int ret;
	launch_worker_thread();
	ret = scsi_init_module();
	if (ret && hpt_worker_thread_handle) {
		DECLARE_MUTEX_LOCKED(sem);
		hpt_notify_wt = &sem;
		stop_thread = 1;
		up(hpt_queue_ready);
		down(&sem);
		hpt_notify_wt = NULL;
	}
	return ret;
}
#endif
EXPORT_NO_SYMBOLS;

/***************************************************************************
 * Function:    void hpt_busy_timer(PChannel pChan)
 * Description: This function will be called when device being busy for a 
				long time.
 */
void hpt_busy_timer(PChannel pChan)
{
    unsigned long		flags;
    spin_lock_irqsave (&io_request_lock, flags);

	AtapiCallBack(pChan);

    spin_unlock_irqrestore (&io_request_lock, flags);
}

/***************************************************************************
 * Function:    void hpt_watch_timer(PChannel pChan)
 * Description: This function will be called when a request waiting for a 
 *     long time. (interrupt may lost)
 */
extern void hpt_watch_timer_Rhpt(PChannel pChan);
void hpt_watch_timer(PChannel pChan)
{
    unsigned long		flags;
    spin_lock_irqsave (&io_request_lock, flags);

	hpt_watch_timer_Rhpt(pChan);

    spin_unlock_irqrestore (&io_request_lock, flags);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
/*
 * proc entry
 */
struct proc_dir_entry proc_scsi_hpt3xx = { 
	0, 							/* inode number - ignore. filled by proc_register_dynamic */
	sizeof(PROC_DIR_NAME)-1, 	/* length of name */
	PROC_DIR_NAME,              /* file name */
	S_IFDIR|S_IRUGO|S_IXUGO,	/* */
	2							/* number of links */
};
#endif

int hpt_set_info(char *buffer,int length)
{
	int retval;
#ifdef SUPPORT_IOCTL
	PUCHAR ke_area;
	int err;
	DWORD dwRet;
	PHPT_IOCTL_PARAM piop=(PHPT_IOCTL_PARAM)buffer;
#endif
	if (length >= 4) {
		if (strncmp(buffer,"hpt ",4) == 0) {
			return (retval = hpt_set_asc_info(buffer+4, length-4)) >= 0? retval : -EINVAL;
		}

#ifdef SUPPORT_IOCTL
        if (piop->Magic == HPT_IOCTL_MAGIC) {
        	KdPrintI(("ioctl=%d in=%x len=%d out=%x len=%d\n", 
        		piop->dwIoControlCode,
        		piop->lpInBuffer,
        		piop->nInBufferSize,
        		piop->lpOutBuffer,
        		piop->nOutBufferSize));
        	if (piop->nInBufferSize) {
				if (verify_area(VERIFY_READ, piop->lpInBuffer, piop->nInBufferSize)) {
					KdPrintE(("Got bad user address.\n"));
					return -EINVAL;
				}
			}
			if (piop->nOutBufferSize) {
				if (verify_area(VERIFY_WRITE, piop->lpOutBuffer, piop->nOutBufferSize)) {
					KdPrintE(("Got bad user address.\n"));
					return -EINVAL;
				}
			}
			if (piop->lpBytesReturned) {
				if (verify_area(VERIFY_WRITE, piop->lpBytesReturned, sizeof(DWORD))) {
					KdPrintE(("Got bad user address.\n"));
					return -EINVAL;
				}
			}
        	if (piop->nInBufferSize+piop->nOutBufferSize > 0x10000) {
        		KdPrintE(("User buffer too large\n"));
        		return -EINVAL;
        	}
        	ke_area = kmalloc(piop->nInBufferSize+piop->nOutBufferSize, GFP_ATOMIC | GFP_DMA);
			if (ke_area == NULL) {
				KdPrintE(("Couldn't allocate kernel mem.\n"));
				return -EINVAL;
			}
			if (piop->nInBufferSize)
				copy_from_user(ke_area, piop->lpInBuffer, piop->nInBufferSize);
			err = Kernel_DeviceIoControl(piop->dwIoControlCode, ke_area, piop->nInBufferSize,
					ke_area + piop->nInBufferSize, piop->nOutBufferSize, &dwRet);    
			if (err==0) {
				if (piop->nOutBufferSize)
					copy_to_user(piop->lpOutBuffer, 
							ke_area + piop->nInBufferSize, 
							piop->nOutBufferSize);
				if (piop->lpBytesReturned)
					copy_to_user(piop->lpBytesReturned, &dwRet, sizeof(DWORD));
			
				kfree(ke_area);
				return length;
			}
			else {
				KdPrintW(("Kernel_ioctl(): return %d\n", err));
			}
			
			kfree(ke_area);
            return -EINVAL;
            
        } else {
            KdPrintW(("Wrong signature: %x\n", piop->Magic));
            return -EINVAL;
        }
#endif // SUPPORT_IOCTL
    }
    
    return -EINVAL;
}

int hpt_copy_info(HPT_GET_INFO *pinfo, char *fmt, ...) 
{
	va_list args;
	char buf[128];
	int len;
  
	va_start(args, fmt);
	len = vsprintf(buf, fmt, args);
	va_end(args);
	hpt_copy_mem_info(pinfo, buf, len);
	return len;
}

void hpt_copy_array_info(HPT_GET_INFO *pinfo, int nld, PVirtualDevice pArray)
{
	int i;
	char *sType, *sStatus=0;
	char arrayname[21];
	char buf[32];
	switch (pArray->arrayType) {
	case VD_RAID_0_STRIPE:
		sType = "RAID 0     ";
		break;
	case VD_RAID_1_MIRROR:
		sType = "RAID 1     ";
		break;
	case VD_RAID_01_2STRIPE:
	case VD_RAID01_MIRROR:
		sType = "RAID 0/1   ";
		break;
	case VD_SPAN:
		sType = "JBOD       ";
		break;
	default:
		sType = "N/A        ";
		break;
	}
	
	if (pArray->arrayType==VD_RAID_01_2STRIPE) {
		int source_bad = RAID_FLAG_ISSET(pArray, RAID_FLAGS_DISABLED|RAID_FLAGS_BROKEN);
		int mirror_bad;
		if (!pArray->pDevice[MIRROR_DISK] ||
			RAID_FLAG_ISSET(pArray->pDevice[MIRROR_DISK]->pArray, RAID_FLAGS_DISABLED|RAID_FLAGS_BROKEN))
			mirror_bad = 1;
		else
			mirror_bad = 0;
		if (source_bad && mirror_bad)
			sStatus = "Disabled";
		else if (source_bad || mirror_bad)
			sStatus = "Critical";
	}
	else if (pArray->RaidFlags & RAID_FLAGS_DISABLED)
		sStatus = "Disabled";
	else if (RAID_FLAG_ISSET(pArray, RAID_FLAGS_BROKEN))
		sStatus = "Critical";
	if (!sStatus) {
		if (pArray->RaidFlags & RAID_FLAGS_BEING_BUILT) {
			sprintf(buf, "Rebuilding %d%%", (pArray->RebuiltSector>>11)*100/(pArray->capacity>>11));
			sStatus = buf;
		}
		else if (pArray->RaidFlags & RAID_FLAGS_NEED_SYNCHRONIZE)
			sStatus = "Critical";
		else
			sStatus = "Normal";
	}
	get_array_name(arrayname, pArray);
	for (i=strlen(arrayname); i<20; i++) arrayname[i] = ' ';
	arrayname[20] = 0;
	hpt_copy_info(pinfo, "%2d  %11s  %s  %5dMB  %s\n", nld, sType, arrayname, pArray->capacity>>11, sStatus);
}

static ST_HPT_DPC DpcQueue[MAX_DPC];
static int DpcQueue_First=0;
static int DpcQueue_Last = 0;

ST_HPT_DPC FastDpcQueue[MAX_DPC];
volatile int FastDpcQueue_First=0;
volatile int FastDpcQueue_Last = 0;

struct semaphore * hpt_notify_wt = 0;
struct semaphore * hpt_queue_ready = 0;
struct task_struct *hpt_worker_thread_handle = 0; // our kernel thread

/****************************************************************
 *	Name:	schedule_command_and_wait
 *	Description:	Start a command and wait for complete. Different 
 *                  from hpt3xx_Command is that it enables kernel scheduler.
 *	Parameters:		SCpnt - Pointer to SCSI command structure.
 *	Returns:		Status code.
 ****************************************************************/
static void hptutil_internal_done (Scsi_Cmnd * SCpnt)
{
	SCpnt->SCp.Status++;
}

static int schedule_command_and_wait(Scsi_Cmnd *SCpnt)
{
	DWORD timeout;
	
    SCpnt->SCp.Status = 0;

	hpt3xx_QueueCommand (SCpnt, hptutil_internal_done);

	timeout = jiffies + 60 * HZ; // should be long than hard-reset time
	while (!SCpnt->SCp.Status && time_before(jiffies, timeout)) schedule();

	if (!SCpnt->SCp.Status) SCpnt->result = (DID_ERROR<<16);
	return SCpnt->result;
}

/***************************************************************************
 * Function:     device_rdwr
 * Description:  Read from / Write to disk/array
 */
BOOL device_rdwr(PDevice pDev, int cmd, ULONG lba, UCHAR nSector, PVOID buffer, DWORD flags)
{
	Scsi_Cmnd SC;
	int i, bus, id;
	PCDB pCdb;
	for (i=0; i<NumAdapters; i++) {
		for (bus=0; bus<hpt_adapters[i]->num_buses; bus++) {
			for (id=0; id<2; id++) {
				if (hpt_adapters[i]->IDEChannel[bus].pDevice[id]==pDev)
					goto found;
			}
		}
	}
	KdPrintE(("Internal error\n"));
	return FALSE;
found:
	ZeroMemory(&SC, sizeof(SC));
	SC.lun = i;
	SC.channel = bus;
	SC.target = id;

	SC.cmd_len = 10;
	SC.cmnd[0] = HPT_IOCTL_CMD;
	pCdb = (PCDB)&SC.cmnd;
	pCdb->CDB10.LogicalBlockByte0 = lba>>24;
	pCdb->CDB10.LogicalBlockByte1 = (lba>>16) & 0xFF;
	pCdb->CDB10.LogicalBlockByte2 = (lba>>8) & 0xFF;
	pCdb->CDB10.LogicalBlockByte3 = lba & 0xFF;
	pCdb->CDB10.TransferBlocksLsb = nSector;
	pCdb->CDB10.TransferBlocksMsb = 0;
			
	SC.sense_buffer[0] = (cmd==IO_COMMAND_READ)? SCSIOP_READ : SCSIOP_WRITE;
	*(PDWORD)&SC.sense_buffer[4] = flags;
	
	SC.SCp.Status = 0;
	SC.request_bufflen = 512 * nSector;
	SC.request_buffer = buffer;

	schedule_command_and_wait(&SC);
	if (SC.result){
		return FALSE;
	}
	return TRUE;
}

/***************************************************************************
 * Function:     hpt_worker_thread
 * Description:  Do background rebuilding. Execute in kernel thread context.
 * Returns:      None
 */
volatile int stop_thread = 0;
void hpt_worker_thread(void *arg)
{
	struct fs_struct *fs;
	DECLARE_MUTEX_LOCKED(semQueue);

	lock_kernel();
	exit_mm(current);

	current->session = 1;
	current->pgrp = 1;

	exit_fs(current);
	fs = init_task.fs;

	if( !fs){
		hpt_printk(("fs is NULL! Is this an SMP kernel on uniproc machine?\n "));
	}
	else{ 
		current->fs = fs;
		atomic_inc(&fs->count);
	}

	/*
	 * Set the name of this process.
	 */
	sprintf(current->comm, "hpt_wt");

	hpt_queue_ready = &semQueue;
	hpt_worker_thread_handle = current;
	
	unlock_kernel();

	if (hpt_notify_wt != NULL) up(hpt_notify_wt);

	while(1) {
		unsigned long flags;
		down(hpt_queue_ready);

		TaskQueue();

		// do one DPC each time
		while (DpcQueue_First!=DpcQueue_Last) {
			ST_HPT_DPC p;
			spin_lock_irqsave(&hptio_request_lock, flags);
			p = DpcQueue[DpcQueue_First];
			DpcQueue_First++;
			DpcQueue_First %= MAX_DPC;
			spin_unlock_irqrestore(&hptio_request_lock, flags);
			
			p.dpc(p.arg);
			schedule();
			/*
			 * we should finish all pending DPCs before terminate the thread.
			 */
			if (stop_thread) {
				PVirtualDevice pArray;
				for (pArray=VirtualDevices; pArray<pLastVD; pArray++) {
					if (pArray->RaidFlags & RAID_FLAGS_BEING_BUILT)
						pArray->RaidFlags |= RAID_FLAGS_ABORT_REBUILD;
				}
			}
		}

		if (stop_thread) break;
	}

	hpt_worker_thread_handle = 0;
	if (hpt_notify_wt != NULL) up(hpt_notify_wt);
}

/***************************************************************************
 * Function:     hpt_queue_dpc
 * Description:  Queue a DPC called by worker thread. It may be called in an ISR.
 * Returns:      None
 */
int hpt_queue_dpc(HPT_DPC dpc, void *arg)
{
	unsigned long flags;
	int p;
	spin_lock_irqsave(&hptio_request_lock, flags);
	p = (DpcQueue_Last + 1) % MAX_DPC;
	if (p==DpcQueue_First) {
	    spin_unlock_irqrestore (&hptio_request_lock, flags);
		KdPrintE(("DPC Queue full!\n"));
		return -1;
	}
	DpcQueue[DpcQueue_Last].dpc = dpc;
	DpcQueue[DpcQueue_Last].arg = arg;
	DpcQueue_Last = p;
	up(hpt_queue_ready);
    spin_unlock_irqrestore (&hptio_request_lock, flags);
	return 0;
}

/***************************************************************************
 * Function:     hpt_queue_fastdpc
 * Description:  Queue a DPC handled by TaskDone.
 * Returns:      None
 ***************************************************************************/
int hpt_queue_fastdpc(HPT_DPC dpc, void *arg)
{
	unsigned long flags;
	int p;	
	spin_lock_irqsave(&hptio_request_lock, flags);
	p = (FastDpcQueue_Last + 1) % MAX_DPC;
	if (p==FastDpcQueue_First) {
	    spin_unlock_irqrestore (&hptio_request_lock, flags);
		KdPrintE(("Fast DPC Queue full!\n"));
		return -1;
	}
	FastDpcQueue[FastDpcQueue_Last].dpc = dpc;
	FastDpcQueue[FastDpcQueue_Last].arg = arg;
	FastDpcQueue_Last = p;
	up(hpt_queue_ready);
    spin_unlock_irqrestore (&hptio_request_lock, flags);
	return 0;
}

void hpt_refresh_logical_devices(void)
{
	Scsi_Device *SDptr;
	int i;
	DWORD inuse = 0;
	unsigned long flags;
	spin_lock_irqsave(&io_request_lock, flags);
	
	/*
	 * first mark all inuse flag as 0 then re-set them
	 */
	for (i=0; i<hpt_num_logical_devices; i++) {
		if (hpt_logical_devices[i].isInUse) inuse |= (1<<i);
		hpt_logical_devices[i].isInUse = 0;
	}

	SDptr = hpt_vhost->host_queue;
	while (SDptr) {
		if (SDptr->id < hpt_num_logical_devices) {
			hpt_logical_devices[SDptr->id].isInUse = SDptr->access_count ? 1 : 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
			if (!hpt_logical_devices[SDptr->id].isValid) SDptr->online = 0;
#endif
		}
		SDptr = SDptr->next;
	}
	
	spin_unlock_irqrestore(&io_request_lock, flags);
}

void hpt_set_device_offline(int id)
{
	Scsi_Device *SDptr = hpt_vhost->host_queue;
	while (SDptr) {
		if (SDptr->id == id) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
			SDptr->online = 0;
#else
			if(id < MAX_DEVICES) hpt_logical_devices[id].isValid = 0;
#endif
			goto out;
		}
		SDptr = SDptr->next;
	}
out:
}

static void TaskStart(PDevice pDev, Scsi_Cmnd *pREQ);

/***************************************************************************
 * Function:     MapRequest
 * Description:  Convert Scsi_Cmnd structure to SCSI_REQUEST_BLOCK
 */
void MapRequest(IN Scsi_Cmnd * pREQ, OUT PSCSI_REQUEST_BLOCK Srb)
{
	ZeroMemory(Srb->SrbExtension, sizeof(SrbExtension));

	Srb->Length 				= sizeof(SCSI_REQUEST_BLOCK);
	Srb->CdbLength				= pREQ->cmd_len;
	memcpy(&Srb->Cdb, &pREQ->cmnd, Srb->CdbLength);

	if (pREQ->cmnd[0]==HPT_IOCTL_CMD) {
		Srb->Function = SRB_FUNCTION_IO_CONTROL;
		Srb->Cdb[0] = pREQ->sense_buffer[0];
		((PSrbExtension)(Srb->SrbExtension))->WorkingFlags = *(PDWORD)&pREQ->sense_buffer[4];
	}
	else {
		Srb->Function = SRB_FUNCTION_EXECUTE_SCSI;
	}
	Srb->SrbStatus				= 0;
	Srb->ScsiStatus				= 0;
	Srb->SenseInfoBufferLength  = 16;

	switch(Srb->Cdb[0]) {
	case SCSIOP_READ6:
	case SCSIOP_READ:
		Srb->SrbFlags = SRB_FLAGS_DATA_IN;
		break;
	case SCSIOP_WRITE6:
	case SCSIOP_WRITE:
		Srb->SrbFlags = SRB_FLAGS_DATA_OUT;
		break;
	default:
		Srb->SrbFlags = 0;
	}

	Srb->DataTransferLength		= pREQ->request_bufflen;
	Srb->TimeOutValue			= 0;
	Srb->DataBuffer				= pREQ->request_buffer;
	Srb->SenseInfoBuffer		= pREQ->sense_buffer;

	if (pREQ->use_sg) {
		((PSrbExtension)(Srb->SrbExtension))->WorkingFlags |= SRB_WFLAGS_USE_SG;
	}
	
	Srb->NextSrb = 0;
}

/***************************************************************************
 * Function:     VOID TaskQueue()
 * Description:  Check if Queue is empty. If there are request in queue,
 *               transfer the request to HIM's request and execute
 *               the request.
 * Returns:      None
 */
VOID TaskQueue()
{
	Scsi_Cmnd * SCpnt;
	PDevice pDev;
	unsigned long flags;

	KdPrintI(("TaskQueue()\n"));
check_next:
	TaskQueue_FDQ();
    
	if (hpt3xx_req_last!=NULL) {
		if (Driver_Busy()) {
			KdPrintI(("Driver busy!\n"));
			goto out;
		}

		spin_lock_irqsave(&queue_request_lock, flags);
		SCpnt = (Scsi_Cmnd *)hpt3xx_req_last->SCp.ptr;
		if (hpt3xx_req_last==SCpnt)
			hpt3xx_req_last = NULL;
		else
			hpt3xx_req_last->SCp.ptr = (char*)SCpnt->SCp.ptr;
		spin_unlock_irqrestore (&queue_request_lock, flags);
	
		/* 
		 * check the command.
		 */
		if (SCpnt->host) {
			if (SCpnt->channel!=0 || SCpnt->target>=hpt_num_logical_devices) {
	bad_target:
				SCpnt->result = (DID_BAD_TARGET<<16);
				Driver_Done();
				SCpnt->scsi_done(SCpnt);
				goto check_next;
			}
			if (!hpt_logical_devices[SCpnt->target].isValid) goto bad_target;
	
			if (hpt_logical_devices[SCpnt->target].isArray) {
				int i;
				PVirtualDevice pArray = (PVirtualDevice)hpt_logical_devices[SCpnt->target].pLD;
				for (i=0; i<MAX_MEMBERS; i++) {
					pDev = pArray->pDevice[i];
					if (pDev) break;
				}
			}
			else {
				pDev = (PDevice)hpt_logical_devices[SCpnt->target].pLD;
				if (pDev->pArray) {
					if (pDev->pArray->arrayType!=VD_RAID_1_MIRROR ||
						pDev!=pDev->pArray->pDevice[0])
						goto bad_target;
					hpt_logical_devices[SCpnt->target].isArray = 1;
					hpt_logical_devices[SCpnt->target].pLD = pDev->pArray;
				}
			}
			if (pDev==0) goto bad_target;
		}
		else {
			if (SCpnt->cmnd[0]!=HPT_IOCTL_CMD)
				goto bad_target;
			if (SCpnt->lun>=NumAdapters || SCpnt->target>1) goto bad_target;
			if (SCpnt->channel >= hpt_adapters[SCpnt->lun]->num_buses) goto bad_target;
			pDev = hpt_adapters[SCpnt->lun]->IDEChannel[SCpnt->channel].pDevice[SCpnt->target];
			if (!pDev) goto bad_target;
		}

		TaskStart(pDev, SCpnt);
		return;
	}
out:
}

/***************************************************************************
 * Function:    int CopySgl(PSCAT_GATH pSG, Scsi_Cmnd *pREQ, int isphy)
 * Description:   Copy SG List
 * Returns:     None
 */
int CopySgl(PSCAT_GATH pSG, Scsi_Cmnd *pREQ, int isphy)
{
	int idx;
	struct scatterlist *sgList = (struct scatterlist *)pREQ->request_buffer;

	if (isphy)
		for (idx = 0; idx < pREQ->use_sg; idx++) {
			pSG[idx].SgAddress = MapToPhysAddr(sgList[idx].address);
			if (pSG[idx].SgAddress & 1) return FALSE;
			pSG[idx].SgSize = sgList[idx].length;
			if (pSG[idx].SgSize & 1) return FALSE;
			pSG[idx].SgFlag = (idx==pREQ->use_sg-1)? SG_FLAG_EOT : 0;
			}
	else
		for (idx = 0; idx < pREQ->use_sg; idx++) {
			pSG[idx].SgAddress = (ULONG)sgList[idx].address;
			pSG[idx].SgSize = sgList[idx].length;
			pSG[idx].SgFlag = (idx==pREQ->use_sg-1)? SG_FLAG_EOT : 0;
			}
	return TRUE;
}

/***************************************************************************
 * Function:    VOID TaskStart(PDevice pDev, Scsi_Cmnd *pREQ)
 * Description: Start a command, doing convert first.
 * Returns:     None
 */
static void TaskStart(PDevice pDev, Scsi_Cmnd *pREQ)
{
	PSCSI_REQUEST_BLOCK Srb;
	PChannel pChan = pDev->pChannel;
	
	KdPrintI(("TaskStart(pDev=%x, pREQ=%x)\n", pDev, pREQ));
	/*
	 * Map Scsi_Cmnd to SRB.
	 */
	Srb=AllocateSrb(pChan);
	Srb->SrbExtension = (PVOID)(((ULONG)&Srb->_SrbExtensionBuffer + 0xF) & ~0xf);
	MapRequest(pREQ, Srb);
	/*
	 * save pChan, pREQ,etc inside SrbExtension
	 */
	((PSrbExtension)(Srb->SrbExtension))->pDev = pDev;
	((PSrbExtension)(Srb->SrbExtension))->pREQ = pREQ;

	AtapiStartIo(pDev, Srb);
}

static void hpt_queue_command_dpc(Scsi_Cmnd *cmd)
{
	hpt3xx_QueueCommand(cmd, cmd->scsi_done);
}

/***************************************************************************
 * Function:    VOID TaskDone(PSCSI_REQUEST_BLOCK Srb)
 * Description: A task execution has been done. For OS request, we need to
 *              Notify OS and invoke next take which wait at queue
 * Returns:     None
 */
VOID TaskDone(PSCSI_REQUEST_BLOCK Srb)
{
	PSrbExtension pSrbExt = (PSrbExtension)Srb->SrbExtension;
	PChannel pChan = pSrbExt->pDev->pChannel;
	Scsi_Cmnd *pREQ = pSrbExt->pREQ;

	switch(SRB_STATUS(Srb->SrbStatus)){
	case SRB_STATUS_SUCCESS:
	case SRB_STATUS_NEED_REQUEUE:
		goto done;

	/*
	 * Some cases we do not retry the request.
	 */
	case SRB_STATUS_INVALID_TARGET_ID:
	case SRB_STATUS_INVALID_PATH_ID:
	case SRB_STATUS_INVALID_LUN:
	case SRB_STATUS_NO_DEVICE:
	case SRB_STATUS_NO_HBA:
	case SRB_STATUS_INVALID_REQUEST:
		KdPrintE(("TaskDone(): SrbStatus=%d (No Retry)\n", Srb->SrbStatus));
		goto done;

	default:
		KdPrintW(("TaskDone(): SrbStatus=%d ScsiStatus=%d\n", Srb->SrbStatus, Srb->ScsiStatus));
		/*
		 * retry the request until MAX_RETRY_COUNT reached.
		 */
		if (pChan->RetryCount++ < MAX_RETRY_COUNT) {
			Srb->SrbStatus = SRB_STATUS_PENDING;
			// re-set pChan->CurrentSrb since it may have been cleared.
			pChan->CurrentSrb = Srb;
			ScheduleRetryProcess(pChan);
			return;
		}
	}

done:
	pChan->pWorkDev = NULL;
	pChan->CurrentSrb = NULL;
	pChan->RetryCount = 0;

	switch(SRB_STATUS(Srb->SrbStatus)) {
	case SRB_STATUS_SUCCESS:
		pREQ->result = (DID_OK<<16);
		break;
		
	case SRB_STATUS_SELECTION_TIMEOUT:
		pREQ->result = (DID_NO_CONNECT<<16);
		break;
		
	case SRB_STATUS_BUSY:
		pREQ->result = (DID_BUS_BUSY<<16);
		break;
		
	case SRB_STATUS_BUS_RESET:
		pREQ->result = (DID_RESET<<16);
		break;
		
	case SRB_STATUS_INVALID_TARGET_ID:
	case SRB_STATUS_INVALID_PATH_ID:
	case SRB_STATUS_INVALID_LUN:
	case SRB_STATUS_NO_HBA:
		pREQ->result = (DID_BAD_TARGET<<16);
		break;
		
	case SRB_STATUS_NO_DEVICE:
		if (pREQ->host)
			hpt_set_device_offline(pREQ->target);
		pREQ->result = (DID_BAD_TARGET<<16);
		break;
		
	case SRB_STATUS_ERROR:
		pREQ->result = (DID_SOFT_ERROR<<16);
		break;
		
	case SRB_STATUS_NEED_REQUEUE:
		 hpt_queue_dpc((HPT_DPC)hpt_queue_command_dpc, pREQ);
		 goto next_req;
		
	default:
		pREQ->result = ((DRIVER_INVALID|SUGGEST_ABORT)<<24) | (DID_ABORT<<16);
	}

	KdPrintI(("TaskDone(pChan=%x, pREQ=%x, result=%x)\n", 
		pChan, pREQ, pREQ->result));
		
	/*
	 * Notify OS that this OS request has been done
	 */
	pREQ->scsi_done (pREQ);
	
next_req:
	Driver_Done();
	TaskQueue();
}

/***************************************************************************
 * Function:    void hptretry(PChannel pChan)
 * Description: This function will be call after waitting 1 ticks ( 1 tick
 *              = 1/Hz second). It will execute retry operation.
 * Returns:     None
 */
void hptretry(PChannel pChan)
{
    unsigned long		flags;

    spin_lock_irqsave (&io_request_lock, flags);

	if (pChan->CurrentSrb) {
		PSrbExtension pSrbExt = (PSrbExtension)pChan->CurrentSrb->SrbExtension;
		KdPrintW(("Retry Request\n"));
		TaskStart(pSrbExt->pDev, pSrbExt->pREQ);
	}
	
    spin_unlock_irqrestore (&io_request_lock, flags);
}

int event_queue_add(PHPT_EVENT pEvent)
{
	int p;
	unsigned long flags;
	spin_lock_irqsave(&hptio_request_lock, flags);	
	p = (event_queue_tail + 1) % MAX_EVENTS;
	if (p==event_queue_head) {
	    spin_unlock_irqrestore (&hptio_request_lock, flags);
		KdPrintE(("Event Queue full!\n"));
		return -1;
	}
	hptdev_event_queue[event_queue_tail] = *pEvent;
	event_queue_tail = p;
    spin_unlock_irqrestore (&hptio_request_lock, flags);
	return 0;
}

int event_queue_remove(PHPT_EVENT pEvent)
{
    unsigned long flags;
	if (event_queue_head != event_queue_tail) {
		spin_lock_irqsave(&hptio_request_lock, flags);
		*pEvent = hptdev_event_queue[event_queue_head];
		event_queue_head++;
		event_queue_head %= MAX_EVENTS;
	    spin_unlock_irqrestore (&hptio_request_lock, flags);
		return 0;
	}
	return -1;
}


void StallExec(ULONG x) {(x < MAX_UDELAY_MS*1000)? udelay(x) : mdelay(x/1000);return;}

void OS_Busy_Handle(PChannel pChan, PDevice pDev)
{
	SetBusyTimer(pChan);
	pDev->DeviceFlags |= DFLAGS_SET_CALL_BACK; 
}

void SetWatchDogTimer(PChannel pChan)
{
	pChan->watch_timer->expires = jiffies + 3*HZ;
	add_timer(pChan->watch_timer);
}
void CancelWatchDogTimer(PChannel pChan) {del_timer(pChan->watch_timer);}

unsigned long spin_lock_irqsave_Rhpt(int lock)
{
	unsigned long flags;
    spin_lock_irqsave((lock == IO_REQUEST_LOCK)?&io_request_lock:
    	(lock == GIO_REQUEST_LOCK)?&gio_request_lock:&hptio_request_lock, flags);
 	return flags;
}

void spin_unlock_irqrestore_Rhpt(int lock, unsigned long flags)
{
	spin_unlock_irqrestore((lock == IO_REQUEST_LOCK)?&io_request_lock:
    	(lock == GIO_REQUEST_LOCK)?&gio_request_lock:&hptio_request_lock, flags);
	return;
}

asmlinkage int printk_Rhpt(const char *fmt, ...)
{
	va_list ap;
	char buf[256];

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	return printk(buf);
}

void outl_Rhpt(unsigned int value, unsigned short port) {outl(value, port); return;}
unsigned int inl_Rhpt(unsigned short port) {return inl(port);}
void outb_Rhpt(unsigned char value, unsigned short port) {outb(value, port); return;}
unsigned char inb_Rhpt(unsigned short port) {return inb(port);}
void outw_Rhpt(unsigned short value, unsigned short port) {outw(value, port); return;}
unsigned short inw_Rhpt(unsigned short port) {return inw(port);}
void insw_Rhpt(unsigned short port, void * addr, unsigned long count) {insw(port, addr, count); return;}
void outsw_Rhpt(unsigned short port, const void * addr, unsigned long count) {outsw(port, addr, count); return;}

unsigned long volatile hpt_get_jiffies(void) {return jiffies;}
int hpt_get_HZ(void) {return HZ;}
void *kmalloc_Rhpt(unsigned int size,int dma) {return kmalloc(size, GFP_ATOMIC | (dma?GFP_DMA:0));}
void kfree_Rhpt(const void *objp) {kfree(objp);}
void schedule_Rhpt(void) {schedule();}
